// -*- C++ -*-
/*!
 * @file OutPortBase.h
 * @brief InPortBase base class
 * @date $Date: 2007-12-31 03:08:05 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2003-2008
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: OutPortBase.h 1754 2010-01-22 07:18:53Z fsi-katami $
 *
 */

#ifndef RTC_OUTPORTBASE_H
#define RTC_OUTPORTBASE_H

#include <vector>
#include <string>

#include <coil/Properties.h>
#include <coil/stringutil.h>

#include <rtm/PortBase.h>
#include <rtm/InPortConsumer.h>
#include <rtm/OutPortProvider.h>
#include <rtm/ConnectorBase.h>
#include <rtm/CdrBufferBase.h>
#include <rtm/SystemLogger.h>
#include <rtm/ConnectorListener.h>

namespace RTC
{
  class PublisherBase;
  class ConnectorBase;
  class OutPortConnector;

  /*!
   * @if jp
   *
   * @class OutPortBase
   *
   * @brief OutPort NX
   * 
   * OutPort ̊NXB
   *
   *
   *
   * Properties: port.outport
   * vpeB
   *
   * - port.outport
   * - port.outport.[name]
   * ConnectorProfile.properties ̏ꍇ
   * - dataport.outport
   *
   * ȉɎw肵̂nB
   * (port.outport.[name]D悳)
   * ɁAꕔ̃vpeB͐ڑ ConnectorProfile ɂ
   * nꍇȀꍇ ConnectorProfile D悳B
   *
   * - input.throughput.profile: enable
   * - input.throughput.update_rate: count [n/count]
   * - input.throughput.total_bytes: [bytes]
   * - input.throughput.total_count: [n]
   * - input.throughput.max_size: [bytes]
   * - input.throughput.min_size: [bytes]
   * - input.throughput.avg_size: [bytes]
   * - input.throughput.byte_sec: [bytes/sec]
   *
   * - output.throughput.profile: enable
   * - output.throughput.update_rate: count [n/count]
   * - output.throughput.total_bytes: [bytes]
   * - output.throughput.total_count:[n]
   * - output.throughput.max_size: [bytes]
   * - output.throughput.min_size: [bytes]
   * - output.throughput.avg_size: [bytes]
   * - output.throughput.max_sendtime: [sec]
   * - output.throughput.min_sendtime: [sec]
   * - output.throughput.avg_sendtime: [sec]
   * - output.throughput.byte_sec: [bytes/sec]
   *
   * dataport.dataflow_type
   * dataport.interface_type
   * dataport.subscription_type
   *
   * [buffer]
   *
   * - buffer.type:
   *     p\ȃobt@̃^Cv
   *     ConnectorProfile ̏ꍇ͗pobt@̃^Cv
   *     w̏ꍇ̓ftHg ringbuffer gpB
   *     ex. ringbuffer, shmbuffer, doublebuffer, etc.
   *     AConsumer, Publisher ̃^CvɂẴ͓obt@^
   *     v邪邽߂́Ȁꍇ͎w͖ƂȂB
   *
   * - buffer.length:
   *     obt@̒
   *
   * - buffer.write.full_policy:
   *     ㏑邩ǂ̃|V[
   *     overwrite (㏑), do_nothing (Ȃ), block (ubN)
   *     block w肵ꍇA timeout lw肷΁Aw莞Ԍ
   *     ݕs\ł΃^CAEgB
   *
   * - buffer.write.timeout:
   *     ^CAEgԂ [sec] Ŏw肷B
   *     1 sec -> 1.0, 1 ms -> 0.001, ^CAEgȂ -> 0.0
   *
   * - buffer.read.empty_policy:
   *     obt@̂Ƃ̓ǂݏo|V[
   *     last (Ō̗vf), do_nothing (Ȃ), block (ubN)
   *     block w肵ꍇA timeout lw肷΁Aw莞Ԍ
   *     ǂݏos\ł΃^CAEgB
   *
   * - buffer.read.timeout:
   *     ^CAEg [sec] Ŏw肷B
   *     1sec -> 1.0, 1ms -> 0.001, ^CAEgȂ -> 0.0
   *
   * - ̑obt@ɌŗLȃIvV
   *
   *
   * [publihser]
   *
   * - publisher.types:
   *      p\ Publisher ̃^Cv
   *      new, periodic, flush, etc..
   *
   * - publisher.push.policy:
   *      InPortփf[^𑗐M|V[
   *      all: obt@ɂ܂Ă邷ׂđM
   *      fifo: obt@FIFOƂ݂ȂđM
   *      skip: Âf[^萔ԈđM
   *      new: ɐVf[^݂̂𑗐M
   *
   * - publisher.push.skip_rate:
   *      push.policy=skip ̂Ƃ̃XLbv
   *      n: nvfɂЂƂM
   *
   * - publisher.periodic.rate:
   *
   * - publisher.thread.type:
   *       Publisher ̃Xbh̃^Cv
   * - publisher.thread.measurement.exec_time: yes/no
   * - publisher.thread.measurement.exec_count: number
   * - publisher.thread.measurement.period_time: yes/no
   * - publisher.thread.measurement.period_count: number
   *
   * [interface]
   *
   * - interface.types:
   *     OutPort interfacẽ^Cv
   *     ex. corba_cdr, corba_any, raw_tcp ȂǃJ}؂ŎwB
   *     w肵ȂΗp\ȂׂẴvoC_gp
   *
   *
   *
   *   
   * OutPort  connect() ł͈ȉ̃V[PXŏsB
   *
   * 1. OutPort Ɋ֘A connector ̐уZbg
   *
   * 2. InPortɊ֘A connector ̎擾
   *  - ConnectorProfile::properties["dataport.corba_any.inport_ref"]
   *    OutPortAny ̃IuWFNgt@Xݒ肳ĂꍇA
   *    t@X擾ConsumerIuWFNgɃZbgB
   *    t@XZbgĂȂΖČpB
   *    (OutPortconnect() ĂяõGg|Cg̏ꍇ́A
   *    InPort̃IuWFNgt@X̓ZbgĂȂ͂łB)
   * 3. PortBase::connect() R[
   *    Port̐ڑ̊{sB
   * 4. L2.InPort̃t@X擾łȂ΁AēxInPort
   *    ֘A connector 擾B
   *
   * 5. ConnectorProfile::properties ŗ^ꂽ񂩂A
   *    OutPort̏sB
   *
   * - [dataport.interface_type]
   * -- CORBA_Any ̏ꍇ: 
   *    InPortAny ʂăf[^B
   *    ConnectorProfile::properties["dataport.corba_any.inport_ref"]
   *    InPortAny ̃IuWFNgt@XZbgB
   * -- RawTCP ̏ꍇ: Raw TCP socket ʂăf[^B
   *    ConnectorProfile::properties["dataport.raw_tcp.server_addr"]
   *    InPort̃T[oAhXZbgB
   *
   * - [dataport.dataflow_type]
   * -- Push̏ꍇ: Subscriber𐶐BSubscriber̃^Cv́A
   *    dataport.subscription_type ɐݒ肳ĂB
   * -- Pull̏ꍇ: InPortf[^Pull^Ŏ擾邽߁A
   *    ɉKvB
   *
   * - [dataport.subscription_type]
   * -- Oncȅꍇ: SubscriberOnce𐶐B
   * -- New̏ꍇ: SubscriberNew𐶐B
   * -- Periodic̏ꍇ: SubscriberPeriodic𐶐B
   *
   * - [dataport.publisher.push_rate]
   * -- dataport.subscription_type=Periodic̏ꍇݒ肷B
   *
   * 6. L̏̂łG[ł΁AG[^[B
   *    ɏsꂽꍇRTC::RTC_OKŃ^[B
   *
   * @since 0.2.0
   *
   * @else
   *
   * @class OutPortBase
   *
   * @brief Output base class.
   *
   * The base class of OutPort<T> which are implementations of OutPort
   *
   * Form a kind of Observer pattern with OutPortBase and PublisherBase.
   * attach(), detach(), notify() of OutPortBase and
   * push() of PublisherBase are methods associated with the Observer pattern.
   *
   * @since 0.2.0
   *
   * @endif
   */
  class OutPortBase
    : public PortBase, public DataPortStatus
  {
  public:
    DATAPORTSTATUS_ENUM

    typedef std::vector<OutPortConnector*> ConnectorList;

    /*!
     * @if jp
     * @brief RXgN^
     *
     * RXgN^B
     *
     * @param name |[g
     *
     * @else
     * @brief Constructor
     *
     * Constructor
     *
     * @param name Port name
     *
     * @endif
     */
    OutPortBase(const char* name, const char* data_type);
    
    /*!
     * @if jp
     * @brief fXgN^
     *
     * fXgN^B
     * o^ꂽSĂ Publisher 폜B
     *
     * @else
     *
     * @brief Destructor
     *
     * Destructor
     * Detach all subscribed Publishers.
     *
     * @endif
     */
    virtual ~OutPortBase(void);
    
    /*!
     * @if jp
     * @brief vpeB̏
     *
     * OutPort̃vpeB
     *
     * @else
     *
     * @brief Initializing properties
     *
     * This operation initializes outport's properties
     *
     * @endif
     */
    void init(coil::Properties& prop);

    /*!
     * @if jp
     *
     * @brief f[^
     *
     * |[gփf[^ށB
     * oChꂽϐɐݒ肳ꂽl|[gɏށB
     *
     * @return ݏ(ݐ:trueAݎs:false)
     *
     * @else
     *
     * @brief Write data
     *
     * Write data to the port.
     * Write the value, which was set to the bound variable, to the port.
     *
     * @return Writing result (Successful:true, Failed:false)
     *
     * @endif
     */
    virtual bool write() = 0;


    /*!
     * @if jp
     * @brief OutPort̂̎擾
     *
     * OutPort̖̂擾B
     *
     * @return |[g
     *
     * @else
     *
     * @brief Retrieve OutPort name
     *
     * Retrieve OutPort name
     *
     * @return the port name
     *
     * @endif
     */
    //    const char* name() const;

    /*!
     * @if jp
     * @brief vpeB擾
     *
     * OutPort̃vpeB擾B
     *
     * @return vpeB
     *
     * @else
     *
     * @brief Get properties
     *
     * Getting properties of this OutPort
     *
     * @return OutPort's properties
     *
     * @endif
     */
    coil::Properties& properties();

    /*!
     * @if jp
     * @brief Connector 擾
     *
     * ݏLĂRlN^擾B
     *
     * @return connector ̃Xg
     *
     * @else
     *
     * @brief Connector list
     *
     * This operation returns connector list
     *
     * @return connector list
     *
     * @endif
     */
    const std::vector<OutPortConnector*>& connectors();

    /*!
     * @if jp
     * @brief ConnectorProfile 擾
     *
     * ݏLĂRlN^Profile擾B
     *
     * @return ConnectorProfile ̃Xg
     *
     * @else
     *
     * @brief ConnectorProfile list
     *
     * This operation returns ConnectorProfile list
     *
     * @return connector list
     *
     * @endif
     */
    ConnectorInfoList getConnectorProfiles();

    /*!
     * @if jp
     * @brief ConnectorId 擾
     *
     * ݏLĂRlN^ID擾B
     *
     * @return ConnectorId ̃Xg
     *
     * @else
     *
     * @brief ConnectorId list
     *
     * This operation returns ConnectorId list
     *
     * @return connector list
     *
     * @endif
     */
    coil::vstring getConnectorIds();

    /*!
     * @if jp
     * @brief Connector̖O擾
     *
     * ݏLĂRlN^̖O擾B
     *
     * @return Connector̃Xg
     *
     * @else
     *
     * @brief Connector name list
     *
     * This operation returns Connector name list
     *
     * @return connector name list
     *
     * @endif
     */
    coil::vstring getConnectorNames();

    /*!
     * @if jp
     * @brief ConnectorProfileIDŎ擾
     *
     * ݏLĂRlN^IDŎ擾B
     *
     * @param id Connector ID
     * @return RlN^ւ̃|C^
     *
     * @else
     *
     * @brief Getting ConnectorProfile by ID
     *
     * This operation returns Connector specified by ID.
     *
     * @param id Connector ID
     * @return A pointer to connector
     *
     * @endif
     */
    OutPortConnector* getConnectorById(const char* id);

     /*!
     * @if jp
     * @brief ConnectorProfile𖼑OŎ擾
     *
     * ݏLĂRlN^𖼑OŎ擾B
     *
     * @param name Connector name
     * @return RlN^ւ̃|C^
     *
     * @else
     *
     * @brief Getting Connector by name
     *
     * This operation returns Connector specified by name.
     *
     * @param id Connector ID
     * @return A pointer to connector
     *
     * @endif
     */
   OutPortConnector* getConnectorByName(const char* name);

    /*!
     * @if jp
     * @brief ConnectorProfileIDŎ擾
     *
     * ݏLĂRlN^IDŎ擾B
     *
     * @param id Connector ID
     * @param prof ConnectorProfile
     * @return false w肵IDȂ
     *
     * @else
     *
     * @brief Getting ConnectorProfile by name
     *
     * This operation returns ConnectorProfile specified by name
     *
     * @param id Connector ID
     * @param prof ConnectorProfile
     * @return false@specified ID does not exist
     *
     * @endif
     */
    bool getConnectorProfileById(const char* id,
                                 ConnectorInfo& prof);

    /*!
     * @if jp
     * @brief ConnectorProfile𖼑OŎ擾
     *
     * ݏLĂRlN^𖼑OŎ擾B
     *
     * @param name Connector name
     * @param prof ConnectorProfile
     * @return false w肵OȂ
     *
     * @else
     *
     * @brief Getting ConnectorProfile by name
     *
     * This operation returns ConnectorProfile specified by name
     *
     * @param id Connector ID
     * @param prof ConnectorProfile
     * @return false specified name does not exist
     *
     * @endif
     */
    bool getConnectorProfileByName(const char* name,
                                   ConnectorInfo& prof);

    
    /*!
     * @if jp
     *
     * @brief InPort activates 
     *
     * InPort activate B
     *
     * @else
     *
     * @brief Activate all Port interfaces
     *
     * This operation activate all interfaces that is registered in the
     * ports.
     *
     * @endif
     */
    virtual void activateInterfaces();

    /*!
     * @if jp
     *
     * @brief SĂ Port ̃C^[tF[X deactivates 
     *
     * Port ɓo^ĂSẴC^[tF[X deactivate B
     *
     * @else
     *
     * @brief Deactivate all Port interfaces
     *
     * This operation deactivate all interfaces that is registered in the
     * ports.
     *
     * @endif
     */
    virtual void deactivateInterfaces();


    /*!
     * @if jp
     * @brief ConnectorDataListener Xiǉ
     *
     * obt@݂܂͓ǂݏoCxgɊ֘Ae탊Xiݒ肷B
     *
     * ݒł郊Xĩ^CvƃR[obNCxg͈ȉ̒ʂ
     *
     * - ON_BUFFER_WRITE:          obt@ݎ
     * - ON_BUFFER_FULL:           obt@t
     * - ON_BUFFER_WRITE_TIMEOUT:  obt@݃^CAEg
     * - ON_BUFFER_OVERWRITE:      obt@㏑
     * - ON_BUFFER_READ:           obt@ǂݏo
     * - ON_SEND:                  InProtւ̑M
     * - ON_RECEIVED:              InProtւ̑M
     * - ON_SEND_ERTIMEOUT:        OutPort^CAEg
     * - ON_SEND_ERERROR:          OutPortG[
     * - ON_RECEIVER_FULL:         InProtobt@t
     * - ON_RECEIVER_TIMEOUT:      InProtobt@^CAEg
     * - ON_RECEIVER_ERROR:        InProtG[
     *
     * Xi ConnectorDataListener pAȉ̃VOj`
     * operator() ĂKvB
     *
     * ConnectorDataListener::
     *         operator()(const ConnectorProfile&, const cdrStream&)
     *
     * ftHgł́Å֐ɗ^XiIuWFNg̏L
     * OutPortɈڂAOutPort̎́A
     * removeConnectorDataListener() ɂ폜ɎIɉ̂B
     * XiIuWFNg̏LĂяoňێꍇ́A3
     *  false w肵AIȉ̂}邱ƂłB
     *
     * @param listener_type Xi^Cv
     * @param listener XiIuWFNgւ̃|C^
     * @param autoclean XiIuWFNg̎Îsǂ̃tO
     *
     * @else
     * @brief Adding BufferDataListener type listener
     *
     * This operation adds certain listeners related to buffer writing and
     * reading events.
     * The following listener types are available.
     *
     * - ON_BUFFER_WRITE:          At the time of buffer write
     * - ON_BUFFER_FULL:           At the time of buffer full
     * - ON_BUFFER_WRITE_TIMEOUT:  At the time of buffer write timeout
     * - ON_BUFFER_OVERWRITE:      At the time of buffer overwrite
     * - ON_BUFFER_READ:           At the time of buffer read
     * - ON_SEND:                  At the time of sending to InPort
     * - ON_RECEIVED:              At the time of finishing sending to InPort
     * - ON_SENDER_TIMEOUT:        At the time of timeout of OutPort
     * - ON_SENDER_ERROR:          At the time of error of OutPort
     * - ON_RECEIVER_FULL:         At the time of bufferfull of InPort
     * - ON_RECEIVER_TIMEOUT:      At the time of timeout of InPort
     * - ON_RECEIVER_ERROR:        At the time of error of InPort
     *
     * Listeners should have the following function operator().
     *
     * ConnectorDataListener::
     *         operator()(const ConnectorProfile&, const cdrStream&)
     *
     * The ownership of the given listener object is transferred to
     * this OutPort object in default.  The given listener object will
     * be destroied automatically in the OutPort's dtor or if the
     * listener is deleted by removeConnectorDataListener() function.
     * If you want to keep ownership of the listener object, give
     * "false" value to 3rd argument to inhibit automatic destruction.
     *
     * @param listener_type A listener type
     * @param listener A pointer to a listener object
     * @param autoclean A flag for automatic listener destruction
     *
     * @endif
     */
    void addConnectorDataListener(ConnectorDataListenerType listener_type,
                                  ConnectorDataListener* listener,
                                  bool autoclean = true);


    /*!
     * @if jp
     * @brief ConnectorDataListener Xi폜
     *
     * ݒ肵e탊Xi폜B
     * 
     * @param listener_type Xi^Cv
     * @param listener XiIuWFNgւ̃|C^
     *
     * @else
     * @brief Removing BufferDataListener type listener
     *
     * This operation removes a specified listener.
     *     
     * @param listener_type A listener type
     * @param listener A pointer to a listener object
     *
     * @endif
     */
    void removeConnectorDataListener(ConnectorDataListenerType listener_type,
                                     ConnectorDataListener* listener);
    

    /*!
     * @if jp
     * @brief ConnectorListener Xiǉ
     *
     * obt@݂܂͓ǂݏoCxgɊ֘Ae탊Xiݒ肷B
     *
     * ݒł郊Xĩ^Cv
     *
     * - ON_BUFFER_EMPTY:       obt@̏ꍇ
     * - ON_BUFFER_READTIMEOUT: obt@Ń^CAEgꍇ
     *
     * Xi͈ȉ̃VOj` operator() ĂKvB
     *
     * ConnectorListener::operator()(const ConnectorProfile&)
     *
     * ftHgł́Å֐ɗ^XiIuWFNg̏L
     * OutPortɈڂAOutPort̎́A
     * removeConnectorListener() ɂ폜ɎIɉ̂B
     * XiIuWFNg̏LĂяoňێꍇ́A3
     *  false w肵AIȉ̂}邱ƂłB
     *
     * @param listener_type Xi^Cv
     * @param listener XiIuWFNgւ̃|C^
     * @param autoclean XiIuWFNg̎Îsǂ̃tO
     *
     * @else
     * @brief Adding ConnectorListener type listener
     *
     * This operation adds certain listeners related to buffer writing and
     * reading events.
     * The following listener types are available.
     *
     * - ON_BUFFER_EMPTY:       At the time of buffer empty
     * - ON_BUFFER_READTIMEOUT: At the time of buffer read timeout
     *
     * Listeners should have the following function operator().
     *
     * ConnectorListener::operator()(const ConnectorProfile&)
     *  
     * The ownership of the given listener object is transferred to
     * this OutPort object in default.  The given listener object will
     * be destroied automatically in the OutPort's dtor or if the
     * listener is deleted by removeConnectorListener() function.
     * If you want to keep ownership of the listener object, give
     * "false" value to 3rd argument to inhibit automatic destruction.
     *
     * @param listener_type A listener type
     * @param listener A pointer to a listener object
     * @param autoclean A flag for automatic listener destruction
     *
     * @endif
     */
    void addConnectorListener(ConnectorListenerType callback_type,
                              ConnectorListener* listener,
                              bool autoclean = true);

    /*!
     * @if jp
     * @brief ConnectorDataListener Xi폜
     *
     * ݒ肵e탊Xi폜B
     * 
     * @param listener_type Xi^Cv
     * @param listener XiIuWFNgւ̃|C^
     *
     * @else
     * @brief Removing BufferDataListener type listener
     *
     * This operation removes a specified listener.
     *     
     * @param listener_type A listener type
     * @param listener A pointer to a listener object
     *
     * @endif
     */
    void removeConnectorListener(ConnectorListenerType callback_type,
                                 ConnectorListener* listener);

    /*!
     * @if jp
     * @brief endian ݒԂ
     *
     * endian ݒboollԂB
     *
     * @return m_littleEndian littlȅꍇtrueAbig̏ꍇfalse ԂB
     *
     * @else
     * @brief
     *
     * return it whether endian setting.
     *
     *@return Return true in the case of "little", false in "big" than it.
     *
     * @endif
     */
    bool isLittleEndian();

    /*!
     * @if jp
     *
     * @brief [CORBA interface] Port ̐ڑs
     *
     * ^ꂽ ConnectoionProfile ̏ɊÂAPortԂ̐ڑm
     * B̊֐͎ɃAvP[VvOc[Ăяo
     * ƂOƂĂB
     * 
     * @param connector_profile ConnectorProfile
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief [CORBA interface] Connect the Port
     *
     * This operation establishes connection according to the given
     * ConnectionProfile inforamtion. This function is premised on
     * calling from mainly application program or tools.
     *
     * @param connector_profile The ConnectorProfile.
     * @return ReturnCode_t The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    connect(ConnectorProfile& connector_profile)
      throw (CORBA::SystemException);


  protected:
    /*!
     * @if jp
     * @brief OutPort̐ݒs
     *
     * properties̏ɊÂOutPort̊eݒs
     *
     * @else
     *
     * @brief Configureing outport
     *
     * This operation configures the outport based on the properties.
     *
     * @endif
     */
    void configure();

    /*!
     * @if jp
     *
     * @brief Interface J
     *
     * ̃Iy[V́Anotify_connect() V[PX̎n߂ɃR[
     * 鏃z֐łB
     * notify_connect() ł́A
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ̏ protected ֐R[ڑsB
     * <br>
     * ̃Iy[V́AVK connector_id ɑ΂Ă͐ڑ̐A
     *  connector_id ɑ΂Ă͍XVK؂ɍsKvB
     *
     * @param connector_profile ڑɊւvt@C
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Publish interface information
     *
     * This operation is pure virutal function that would be called at the
     * beginning of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile() 
     *
     * This operation should create the new connection for the new
     * connector_id, and should update the connection for the existing
     * connection_id.
     *
     * @param connector_profile The connection profile information
     *
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    publishInterfaces(ConnectorProfile& connector_profile);
    
    /*! @if jp
     *
     * @brief Interface ɐڑ
     *
     * ̃Iy[V́Anotify_connect() V[PX̒ԂɃR[
     * 鏃z֐łB
     * notify_connect() ł́A
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ̏ protected ֐R[ڑsB
     *
     * @param connector_profile ڑɊւvt@C
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Subscribe to the interface
     *
     * This operation is pure virutal function that would be called at the
     * middle of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * @param connector_profile The connection profile information
     *
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    subscribeInterfaces(const ConnectorProfile& connector_profile);
    
    /*!
     * @if jp
     *
     * @brief Interface ̐ڑ
     *
     * ̃Iy[V́Anotify_disconnect() V[PX̏IɃR[
     * 鏃z֐łB
     * notify_disconnect() ł́A
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile()
     * ̏ protected ֐R[ڑsB
     *
     * @param connector_profile ڑɊւvt@C
     *
     * @else
     *
     * @brief Disconnect the interface connection
     *
     * This operation is pure virutal function that would be called at the
     * end of the notify_disconnect() process sequence.
     * In the notify_disconnect(), the following methods would be called.
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile() 
     *
     * @param connector_profile The profile information associated with 
     *                          the connection
     *
     * @endif
     */
    virtual void
    unsubscribeInterfaces(const ConnectorProfile& connector_profile);

    /*!
     * @if jp
     * @brief OutPort provider ̏
     * @else
     * @brief OutPort provider initialization
     * @endif
     */
    void initProviders();

    /*!
     * @if jp
     * @brief InPort consumer ̏
     * @else
     * @brief InPort consumer initialization
     * @endif
     */
    void initConsumers();

    bool checkEndian(const coil::Properties& prop, bool& littleEndian);

    /*!
     * @if jp
     * @brief OutPort provider ̐
     * @else
     * @brief OutPort provider creation
     * @endif
     */
    OutPortProvider* createProvider(ConnectorProfile& cprof,
                                    coil::Properties& prop);
    /*!
     * @if jp
     * @brief InPort consumer ̐
     * @else
     * @brief InPort consumer creation
     * @endif
     */
    InPortConsumer* createConsumer(const ConnectorProfile& cprof,
                                   coil::Properties& prop);
    
    /*!
     * @if jp
     * @brief OutPortPushConnector ̐
     * @else
     * @brief OutPortPushConnector creation
     * @endif
     */
    OutPortConnector* createConnector(const ConnectorProfile& cprof,
                                      coil::Properties& prop,
                                      InPortConsumer* consumer);
    /*!
     * @if jp
     * @brief OutPortPullConnector ̐
     * @else
     * @brief OutPortPullConnector creation
     * @endif
     */
    OutPortConnector* createConnector(const ConnectorProfile& cprof,
                                      coil::Properties& prop,
                                      OutPortProvider* provider);

  protected:
    /*!
     * @if jp
     * @brief vpeB
     * @else
     * @brief Properties
     * @endif
     */
    coil::Properties m_properties;
    /*!
     * @if jp
     * @brief ڑXg
     * @else
     * @brief Connection list
     * @endif
     */
    std::vector<OutPortConnector*> m_connectors;
    /*!
     * @if jp
     * @brief p\provider
     * @else
     * @brief Available providers
     * @endif
     */
    coil::vstring m_providerTypes;
    /*!
     * @if jp
     * @brief p\consumer
     * @else
     * @brief Available consumers
     * @endif
     */
    coil::vstring m_consumerTypes;
    /*!
     * @if jp
     * @brief ڑGfBA 
     * @else
     * @brief Connected Endian
     * @endif
     */
    bool m_littleEndian;
    /*!
     * @if jp
     * @brief ConnectorDataListener Xi
     * @else
     * @brief ConnectorDataListener listener
     * @endif
     */
    ConnectorListeners m_listeners;
   
    /*!
     * @if jp
     * @brief provider 폜邽߂ Functor
     * @else
     * @brief Functor to delete providers
     * @endif
     */
    struct provider_cleanup;

    /*!
     * @if jp
     * @brief connector 폜邽߂ Functor
     * @else
     * @brief Functor to delete connectors
     * @endif
     */
    struct connector_cleanup;
  };
}; // End of namespace RTC
#endif // RTC_RTCOUTPORTBASE_H
